# coding=utf-8

"""Utilities for Mercurial emulator."""

import os
import os.path as osp
try:
    from subprocess import PIPE, run
except ImportError:
    # This allows to execute few tasks on calibre9 (with python 3.4)
    from subprocess import PIPE, Popen

    def run(*args, **kwargs):
        """Minimal replacement of 'subprocess.run' for python<3.5"""
        class ProcEmulator(object):
            """Return of 'run'."""
            returncode = stdout = stderr = None

        proc = ProcEmulator()
        subp = Popen(*args, **kwargs)
        proc.stdout, proc.stderr = subp.communicate()
        proc.returncode = subp.returncode
        return proc

from aslint.logger import logger
from aslint.string_utils import convert


def is_repository(path):
    """Tell if `path` contains a Mercurial repository."""
    return osp.isdir(osp.join(path, ".hg"))

def hgcmd(cmd, repopath, *args, **opts):
    """Return the output of `cmd`.

    Returns:
        (int, str): Exit code and output of the command.
    """
    if 'message' in opts:
        opts['message'] = repr(convert(opts['message']))

    if type(cmd) is tuple:
        cmd = list(cmd)
    elif type(cmd) is not list:
        cmd = [cmd]
    cmd.extend(args)
    for key, value in list(opts.items()):
        if value in (None, False):
            continue
        if value is True:
            cmd.append("--{0}".format(key))
        else:
            cmd.append("--{0}={1}".format(key, value))

    return shell_cmd(repopath, cmd)

def shell_cmd(path, cmd):
    """Return the output of `cmd` using an external shell."""
    if type(cmd) is tuple:
        cmd = list(cmd)
    elif type(cmd) is not list:
        cmd = [cmd]
    cmd.insert(0, 'hg')
    if path:
        cmd.insert(1, '-R')
        cmd.insert(2, path)
    logger.debug(cmd)
    proc = run(cmd, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    return proc.returncode, proc.stdout.strip()

def get_rev_log(repopath, rev='0', template='{node}\n'):
    """Return the hash of the revision 'rev'."""
    opts = {
        'rev': rev,
        'template': template,
    }
    try:
        out = hgcmd("log", repopath, **opts)[1]
    except Exception:
        out = None
    return out
